
# synopsis:
#   Search XML documents for regex pattern.
#   Use xpath.
# usage:
#   ./search_xml.exs [opts] regex-pattern file1 file2 ...
# options:
#   --verbose       Print extra information.  Default: false.
#   --fields="xxx"  Search in fields.  Possible fields are: tag, text,
#                   attr-name, attr-value.  Example: "tag:text:attr-value".
#                   Default: "".
# examples:
#   ./search_xml.exs --verbose pattern file1 file2 ...
#   ./search_xml.exs --verbose --fields="tag:text" "a-name" file1 file2 ...


defmodule SearchXml do

  require IEx
  require Record
  Record.defrecord :xmlElement,
    Record.extract(:xmlElement, from_lib: "xmerl/include/xmerl.hrl")
  Record.defrecord :xmlText,
    Record.extract(:xmlText, from_lib: "xmerl/include/xmerl.hrl")
  Record.defrecord :xmlAttribute,
    Record.extract(:xmlAttribute, from_lib: "xmerl/include/xmerl.hrl")

  #
  # Return all nodes that have the specified tag.
  @spec search_file_tag(Tuple.t, Regex.t, String.t) :: List.t(Tuple.t)
  defp search_file_tag(root, re_pattern, infilename) do
    to_string_fn = fn (x) -> to_string(x) end
    show_fn = fn(node1) ->
      content = xmlElement(node1, :name)
      parents = xmlElement(node1, :parents)
      |> Keyword.keys()
      |> Enum.map(to_string_fn)
      |> Enum.reverse()
      |> Enum.join("/")
      IO.puts("#{infilename}::#{parents}::#{content} (tag)")
      node1
    end
    match_fn = fn (node1) ->
      case elem(node1, 0) do
        :xmlElement ->
          tag = to_string(xmlElement(node1, :name))
          Regex.match?(re_pattern, tag)
        _ ->
          false
      end
    end
    matched_nodes = :xmerl_xpath.string('.//node()', root)
    |> Enum.filter(match_fn)
    |> Enum.map(show_fn)
    #IEx.pry
    matched_nodes
  end # search_file_tag

  #
  # Return all nodes that have the specified attribute name.
  @spec search_file_attrib(Tuple.t, Regex.t, String.t) :: List.t(Tuple.t)
  defp search_file_attrib(root, re_pattern, infilename) do
    to_string_fn = fn (x) -> to_string(x) end
    show_fn = fn(node1) ->
      content = xmlAttribute(node1, :value)
      parents = xmlAttribute(node1, :parents)
      |> Keyword.keys()
      |> Enum.map(to_string_fn)
      |> Enum.reverse()
      |> Enum.join("/")
      IO.puts("#{infilename}::#{parents}::#{content} (attribute)")
      node1
    end
    match_fn = fn (node1) ->
      case elem(node1, 0) do
        :xmlAttribute ->
          attr_value = to_string(xmlAttribute(node1, :value))
          Regex.match?(re_pattern, attr_value)
        _ ->
          false
      end
    end
    matched_nodes = :xmerl_xpath.string('.//@*', root)
    |> Enum.filter(match_fn)
    |> Enum.map(show_fn)
    matched_nodes
  end # search_file_attrib

  #
  # Return all text nodes that contain the specified regex pattern.
  @spec search_file_text(Tuple.t, Regex.t, String.t) :: List.t(Tuple.t)
  defp search_file_text(root, re_pattern, infilename) do
    match_fn = fn (node1) ->
      text = xmlText(node1, :value)
      text1 = to_string(text)
      Regex.match?(re_pattern, text1)
    end
    to_string_fn = fn (x) -> to_string(x) end
    show_fn = fn(node1) ->
      content = xmlText(node1, :value)
      #IEx.pry
      parents = xmlText(node1, :parents)
      |> Keyword.keys()
      |> Enum.map(to_string_fn)
      |> Enum.reverse()
      |> Enum.join("/")
      IO.puts("#{infilename}::#{parents}::#{content} (text)")
      node1
    end
    pattern = './/text()'
    nodes = :xmerl_xpath.string(pattern, root)
    matched_nodes = Enum.filter(nodes, match_fn)
    Enum.each(matched_nodes, show_fn)
    matched_nodes = :xmerl_xpath.string(pattern, root)
    |> Enum.filter(match_fn)
    |> Enum.map(show_fn)
    matched_nodes
  end # search_file_text

  #
  # Perform searchs on a single file.
  # Search the text, the tags, and the attributes.
  @spec search_file(Tuple.t, Regex.t, Keyword.t, String.t) :: List.t(Tuple.t)
  defp search_file(root, re_pattern, fields, infilename) do
    text_nodes = if Keyword.get(fields, :text) do
      search_file_text(root, re_pattern, infilename)
    end
    tag_nodes = if Keyword.get(fields, :tag) do
      search_file_tag(root, re_pattern, infilename)
    end
    attr_nodes = if Keyword.get(fields, :attribute) do
      search_file_attrib(root, re_pattern, infilename)
    end
    nodes = List.flatten([text_nodes, tag_nodes, attr_nodes])
    #IEx.pry
    nodes
  end # search_file

  @doc """
  Search each of a list of files for pattern.
  The options argument contains the fields to be searched.
  """
  @spec search(Keyword.t, String.t, List.t(String.t)) :: nil
  def search(options, pattern, infilenames) do
    #IO.inspect(options)
    _verbose = Keyword.get(options, :verbose, false)
    fieldstrs = Keyword.get(options, :fields, "")
    field_fn = fn (val) -> {String.to_atom(val), true} end
    fields = Enum.map(String.split(fieldstrs, ":"), field_fn)
    #IO.inspect(fields)
    {:ok, re_pattern} = Regex.compile(pattern)
    fn_search = fn (infilename) ->
      #sep = "--------------------------"
      #IO.puts("#{sep}\nDoc: #{infilename}\n#{sep}")
      {root, []} = :xmerl_scan.file(infilename)
      search_file(root, re_pattern, fields, infilename)
    end # fn
    Enum.each(infilenames, fn_search)
  end # search

  # Print a usage message.
  # If variable `invalid` is not empty, print an error message.
  @spec usage(List.t(Tuple.t)) :: nil
  defp usage(invalid) do
    usage_message = """
    synopsis:
      Search XML documents for regex pattern.
      Use xpath.
    usage:
      ./search_xml.exs [opts] regex-pattern file1 file2 ...
    options:
      --verbose       Print extra information.  Default: false.
      --fields="xxx"  Search in fields.  Possible fields are: tag, text,
                      attr-name, attr-value.  Example: "tag:text:attr-value".
                      Default: "".
    examples:
      ./search_xml.exs --verbose pattern file1 file2 ...
      ./search_xml.exs --verbose --fields="tag:text" "a-name" file1 file2 ...
    """
    IO.puts(:stderr, usage_message)
    unless invalid == [] do
        IO.puts(:stderr, "invalid options:")
        IO.inspect(:stderr, invalid, [])
    end # unless
  end # usage

  @doc """
  Parse command line, then start the search.
  If the command line has an error or "--help", display usage.
  """
  @spec main() :: nil
  def main() do
    opt_spec = [help: :boolean, verbose: :boolean, fields: :list]
    {options, args, invalid} = OptionParser.parse(System.argv(), opt_spec)
    case invalid do
      [] ->
        help = Keyword.get(options, :help, false)
        case help do
          true ->
            usage([])
          false ->
            case args do
              [] ->
                usage([])
              [pattern | infilenames] ->
                search(options, pattern, infilenames)
            end
        end
      _ ->
        usage(invalid)
    end
  end # end main

  def test() do
    opt = [fields: "tag:text:attribute"]
    pattern = "(Sammy)|(a00)|(artist)|(01)"
    #files = ["albums03.xml", "albums02.xml"]
    files = ["albums03.xml", ]
    search(opt, pattern, files)
  end # test

end # module SearchXml
